use cairo;
use chrono;
use memfd::{self, Memfd};
use memmap2::MmapMut;
use pangocairo;
use std::hash::{DefaultHasher, Hash, Hasher};
use std::os::fd::AsFd;

use wayland_client::protocol::wl_shm;

use crate::wayland_protocol::State;

impl State {
    pub fn draw(&mut self) {
        if self.debug {
            eprintln!("draw");
        }
        let (init_w, init_h) = (
            self.width.unwrap() * self.scale,
            self.height.unwrap() * self.scale,
        );

        let fd: Memfd = memfd::MemfdOptions::default()
            .create("peanutbutter")
            .unwrap();
        let file = fd.as_file();
        let cairo_fmt = cairo::Format::ARgb32;
        let stride = cairo_fmt.stride_for_width(init_w as u32).unwrap();
        file.set_len((stride * init_h) as u64).unwrap();
        let mmap: MmapMut = unsafe { MmapMut::map_mut(file).unwrap() };

        let surface = cairo::ImageSurface::create_for_data(
            mmap,
            cairo::Format::ARgb32,
            init_w,
            init_h,
            stride,
        )
        .unwrap();
        let context = cairo::Context::new(&surface).unwrap();
        if self.touch_state.is_some() && self.touch_state.as_ref().unwrap().permalocked() {
            context.set_source_rgb(1.0, 0.0, 0.0);
        } else if self.inverted {
            context.set_source_rgb(1.0, 1.0, 1.0);
        } else {
            context.set_source_rgb(0.0, 0.0, 0.0);
        }
        // paint background colour
        context.paint().unwrap();
        if self.inverted {
            context.set_source_rgb(0.0, 0.0, 0.0);
        } else {
            context.set_source_rgb(1.0, 1.0, 1.0);
        }

        // draw status line
        let status = self.status.lock().expect("status unlock must work");
        let layout = pangocairo::functions::create_layout(&context);
        layout.set_font_description(Some(&self.fontdesc_status));
        layout.set_markup(&status);
        let (layout_w, _) = layout.pixel_size();
        context.move_to(
            (init_w as f64 / 2.0) - (layout_w as f64 / 2.0),
            (50 * self.scale) as f64,
        );
        pangocairo::functions::show_layout(&context, &layout);

        // hash the statusline so we can check if it changes later
        let mut h = DefaultHasher::new();
        status.hash(&mut h);
        self.statushash = Some(h.finish());

        //draw time
        self.time = Some(chrono::Local::now());
        let timestring = self.time.as_ref().unwrap().format("%H:%M").to_string();
        let layout = pangocairo::functions::create_layout(&context);
        layout.set_font_description(Some(&self.fontdesc_time));
        layout.set_text(&timestring);
        let (layout_w, layout_h) = layout.pixel_size();
        context.move_to(
            (init_w as f64 / 2.0) - (layout_w as f64 / 2.0),
            (init_h as f64 / 2.0) - (layout_h as f64 / 2.0),
        );
        pangocairo::functions::show_layout(&context, &layout);

        let pool = self.wl_shm.as_ref().unwrap().create_pool(
            file.as_fd(),
            (init_w * init_h * 4) as i32,
            &self.queue_handle,
            (),
        );
        let buffer = pool.create_buffer(
            0,
            init_w as i32,
            init_h as i32,
            (init_w * 4) as i32,
            wl_shm::Format::Argb8888,
            &self.queue_handle,
            (),
        );

        self.buffers.push_back(buffer.clone());

        if self.configured {
            let surface = self.base_surface.as_ref().unwrap();
            surface.set_buffer_scale(self.scale);
            surface.frame(&self.queue_handle, surface.clone());
            surface.attach(Some(&buffer), 0, 0);
            surface.damage(0, 0, i32::MAX, i32::MAX);
            surface.commit();
        }

        while self.buffers.len() >= 3 {
            let oldbuffer = self.buffers.pop_front().unwrap();
            oldbuffer.destroy();
        }

        pool.destroy();
    }
}
