use std::hash::{DefaultHasher, Hash, Hasher};
use std::io::{BufRead, BufReader};
use std::process::{Command, Stdio};
use std::sync::{Arc, Mutex};
use std::{env, thread, time};

mod draw;
mod poller;
mod wayland_protocol;
use chrono::Timelike;
use poller::Poller;
use wayland_protocol::init_wayland;
use wayland_protocol::TouchDown;

mod touch_pass;
use touch_pass::TouchState;

#[derive(Copy, Clone, PartialEq)]
enum ParseArg {
    None,
    Font,
    StatusCommand,
    Tries,
    DeepLockTime,
}

// The main function of our program
fn main() {
    let mut debug = false;
    let mut notouch = false;
    let mut nofeedback = false;
    let mut statuscommand: Option<Vec<String>> = None;
    let mut parsearg = ParseArg::None;
    let mut resolution = 2;
    let mut tries: u16 = 10;
    let mut deeplocktime = time::Duration::from_secs(20);
    let mut paranoid = false;
    let mut fontname: String = "Sans".into();
    for argument in env::args().skip(1) {
        match argument.as_str() {
            "-d" | "--debug" => debug = true,
            "-T" | "--notouch" => notouch = true,
            "-F" | "--nofeedback" => nofeedback = true,
            "-s" | "--statuscommand" => {
                statuscommand = Some(Vec::new());
                parsearg = ParseArg::StatusCommand;
            }
            "-f" | "--font" => parsearg = ParseArg::Font,
            "-t" | "--tries" => parsearg = ParseArg::Tries,
            "-L" | "--deeplock" => parsearg = ParseArg::DeepLockTime,
            "-p" | "--paranoid" => paranoid = true,
            "-1" => resolution = 1,
            "-2" => resolution = 2,
            "-3" => resolution = 3,
            "-4" => resolution = 4,
            "-h" | "--help" => {
                eprintln!("Options:");
                eprintln!(
                    " -d / --debug                    Debug mode (allows quiting on any keypress) and prints verbose logging (insecure!)"
                );
                eprintln!(" -T / --notouch                  Disable touch support");
                eprintln!(" -F / --nofeedback               Disable touch feedback");
                eprintln!(" -f / --font [font               Set font family");
                eprintln!(" -t / --tries [tries]                 The number of incorrect touches allowed before the locker goes into a deeper lock (0 = unlimited, default = {})", tries);
                eprintln!(" -L / --deeplock [seconds]            Timeout in seconds for deeplock mode, during which the device can not be unlocked. (default = {})", deeplocktime.as_secs());
                eprintln!(" -p / --paranoid                      In paranoid mode, the deeplock timeout will be doubled every time it is reached, making successive attempts harder and harder.");
                eprintln!(" -1                              1x1 touch grid (insecure)");
                eprintln!(
                    " -2                              2x2 touch grid (1=topleft, 4=bottomright)"
                );
                eprintln!(
                    " -3                              3x3 touch grid (1=topleft, 9=bottomright)"
                );
                eprintln!(
                    " -4                              4x4 touch grid (0=topleft, f=bottomright)"
                );
                eprintln!(" -s / --statuscommand [command]  Run this command for a status line, it should run continuously and each new output line will become the status line. This parameter should be the last parameter provided. Any further parameters will be passed to the command.");
                std::process::exit(0);
            }
            "-v" | "--version" => {
                println!(env!("CARGO_PKG_VERSION"));
                std::process::exit(0);
            }
            _ => match parsearg {
                ParseArg::Font => {
                    fontname = argument;
                    parsearg = ParseArg::None;
                }
                ParseArg::StatusCommand => {
                    if let Some(statuscommand) = statuscommand.as_mut() {
                        statuscommand.push(argument);
                    }
                }
                ParseArg::Tries => {
                    tries = argument.parse().expect("Tries must be an integer");
                }
                ParseArg::DeepLockTime => {
                    deeplocktime = time::Duration::from_secs(
                        argument
                            .parse()
                            .expect("Deep lock time must be an integer (seconds)"),
                    );
                }
                ParseArg::None => {
                    eprintln!("invalid option: {}", argument);
                    std::process::exit(1);
                }
            },
        }
    }

    let status: Arc<Mutex<String>> = Arc::new(Mutex::new(String::new()));

    if let Some(statuscommand) = statuscommand {
        let status_shared = status.clone(); //this one goes into the thread
        thread::spawn(move || {
            //spawn status process and read output (in separate thread)
            let mut iter = statuscommand.into_iter();
            match Command::new(
                iter.next()
                    .expect("expected a command after --statuscommand"),
            )
            .args(iter)
            .stdout(Stdio::piped())
            .spawn()
            {
                Ok(child) => {
                    if let Some(command_output) = child.stdout {
                        let reader = BufReader::new(command_output);
                        for line in reader.lines() {
                            match line {
                                Ok(line) => {
                                    let mut status_shared =
                                        status_shared.lock().expect("unlock must work");
                                    *status_shared = line;
                                }
                                Err(e) => {
                                    eprintln!("Unable to obtain status line: {:?}", e);
                                    break;
                                }
                            }
                        }
                    }
                }
                Err(e) => {
                    eprintln!("Failed to launch statuscommand: {}", e);
                }
            }
        });
    }

    let passcode = if let Ok(passcode) = env::var("PEANUTBUTTER_PASSCODE") {
        Some(passcode)
    } else {
        None
    };

    let (mut way_state, mut event_queue) =
        init_wayland(&fontname, debug, notouch, nofeedback, status);
    let wait_between_lock_unlock = time::Duration::from_secs(2);
    let wait_between_polling = time::Duration::from_millis(250);
    let mut reset_touchstate: Option<chrono::DateTime<chrono::Local>> = None;

    way_state.draw(); //initial frame
    way_state.lock();
    let mut previous_down: Option<TouchDown> = None;
    way_state.touch_state = Some(TouchState::new(
        way_state.width.unwrap(),
        way_state.height.unwrap(),
        passcode.as_deref(),
        resolution,
        tries,
    ));

    //single blocking dispatch before the main loop
    event_queue.blocking_dispatch(&mut way_state).unwrap();

    let mut poller = Poller::new(&way_state, 1000);

    while way_state.is_running() {
        // flush the outgoing buffers to ensure that the server does receive the messages
        // you've sent
        event_queue.flush().expect("event queue flush");

        // This puts in place some internal synchronization to prepare for the fact that
        // you're going to wait for events on the socket and read them, in case other threads
        // are doing the same thing
        let read_guard = event_queue.prepare_read().expect("read event queue");

        // Poll for events (non-blocking)
        if let Some((wayland_socket_ready, timer_ready)) = poller.poll() {
            if wayland_socket_ready {
                // If epoll notified readiness of the Wayland socket, you can now proceed to the read
                read_guard.read().expect("read guard must work");
                // And now, you must invoke dispatch_pending() to actually process the events
                event_queue
                    .dispatch_pending(&mut way_state)
                    .expect("dispatch pending must work");
            } else {
                // otherwise, some of your other FD are ready, but you didn't receive Wayland events,
                // you can drop the guard to cancel the read preparation
                std::mem::drop(read_guard);
            }

            if timer_ready {
                poller.read_timer();
                let time = chrono::Local::now();
                let time_changed = way_state.time.is_none()
                    || way_state.time.as_ref().unwrap().minute() != time.minute();
                let status_changed = {
                    let status = way_state.status.lock().expect("unlock must work");
                    let mut h = DefaultHasher::new();
                    status.hash(&mut h);
                    let statushash = h.finish();
                    way_state.statushash.is_none()
                        || *way_state.statushash.as_ref().unwrap() != statushash
                };
                if way_state.debug {
                    if time_changed {
                        eprintln!("time changed");
                    }
                    if status_changed {
                        eprintln!("status changed");
                    }
                }
                if let Some(reset_time) = reset_touchstate.as_ref() {
                    //we're in a deep lock state, check if we can exit again
                    if time >= *reset_time {
                        if way_state.debug {
                            eprintln!("deeplock state timed-out");
                        }
                        way_state.touch_state.as_mut().unwrap().reset();
                        way_state.redraw = true;
                        reset_touchstate = None;
                    }
                }
                if time_changed || status_changed {
                    way_state.redraw = true;
                }
            }

            if let Some(last_down) = way_state.last_down.as_ref() {
                if previous_down.as_ref().is_none()
                    || previous_down.as_ref().unwrap().serial != last_down.serial
                {
                    previous_down = Some(*last_down);
                    if let Some(touch_state) = way_state.touch_state.as_mut() {
                        let code = touch_state.add_touch(last_down.x, last_down.y);
                        if way_state.debug {
                            eprintln!("touch: {}", code);
                        }
                        if touch_state.is_pass_valid() {
                            if way_state.debug {
                                eprintln!("code accepted");
                            }
                            break; //exit main loop
                        }
                        touch_state.fail();
                        if reset_touchstate.is_none() && touch_state.permalocked() {
                            //we entered a deep lock timeout stage because the number of tries exceeded the allowed number
                            //set a timer when it expires again
                            reset_touchstate = Some(chrono::Local::now() + deeplocktime);
                            if way_state.debug {
                                eprintln!("entering deeplock state until {:?}", reset_touchstate);
                            }
                            if paranoid {
                                deeplocktime = deeplocktime * 2;
                            }
                        }
                        if !way_state.nofeedback {
                            way_state.invert();
                            way_state.redraw = true;
                        }
                    }
                }
            }

            if way_state.redraw {
                if let Some(surface) = way_state.base_surface.as_ref() {
                    surface.frame(&way_state.queue_handle, surface.clone());
                    surface.commit();
                }
            }
        } else {
            thread::sleep(wait_between_polling);
        }
    }
    way_state.unlock();
    thread::sleep(wait_between_lock_unlock);
}
