use libc::{epoll_event, epoll_wait, EPOLL_CLOEXEC, EPOLL_CTL_ADD};
use std::os::fd::{AsFd, AsRawFd};
use std::time::Duration;
use timerfd::{SetTimeFlags, TimerFd, TimerState};

use crate::wayland_protocol::State;

const MAX_EVENTS: usize = 16;

/// This structure abstracts over specific file descriptors for the main event loop
pub struct Poller {
    epoll_fd: i32,
    wayland_fd: i32,
    timer: TimerFd,
    events: [epoll_event; MAX_EVENTS],
}

impl Poller {
    pub fn new(state: &State, timer_interval_ms: u32) -> Self {
        let epoll_fd = unsafe { libc::epoll_create1(EPOLL_CLOEXEC) };
        assert!(epoll_fd >= 0);
        let wayland_fd = Self::add_wayland_fd(state, epoll_fd);
        let timer = Self::add_timer_fd(timer_interval_ms, epoll_fd);
        let events = [epoll_event { events: 0, u64: 0 }; MAX_EVENTS];
        Self {
            epoll_fd,
            wayland_fd,
            timer,
            events,
        }
    }

    fn add_timer_fd(interval_ms: u32, epoll_fd: i32) -> TimerFd {
        let mut timer = TimerFd::new().unwrap();
        let interval = Duration::new(0, interval_ms * 1000000);
        timer.set_state(
            TimerState::Periodic {
                current: interval.clone(),
                interval,
            },
            SetTimeFlags::Default,
        );
        let mut timer_event = libc::epoll_event {
            events: libc::EPOLLIN as u32,
            u64: timer.as_raw_fd() as u64,
        };
        let ret = unsafe {
            libc::epoll_ctl(epoll_fd, EPOLL_CTL_ADD, timer.as_raw_fd(), &mut timer_event)
        };
        assert!(ret >= 0);
        timer
    }

    fn add_wayland_fd(state: &State, epoll_fd: i32) -> i32 {
        let wayland_fd = state.connection.as_fd().as_raw_fd();
        let mut wayland_event = libc::epoll_event {
            events: libc::EPOLLIN as u32,
            u64: wayland_fd as u64,
        };
        let ret =
            unsafe { libc::epoll_ctl(epoll_fd, EPOLL_CTL_ADD, wayland_fd, &mut wayland_event) };
        assert!(ret >= 0);
        wayland_fd
    }

    /// Non-blocking polling, returns whether wayland and/or the timer is ready
    pub fn poll(&mut self) -> Option<(bool, bool)> {
        let mut wayland_socket_ready = false;
        let mut timer_ready = false;
        let timer_fd = self.timer.as_raw_fd();
        unsafe {
            let num_event = epoll_wait(
                self.epoll_fd,
                self.events.as_mut_ptr(),
                MAX_EVENTS as i32,
                1000,
            );
            if num_event == 0 {
                return None;
            } else if num_event < 0 {
                eprintln!("[peanutbutter] epoll_wait() returned error {}", num_event);
            }
            for e in self.events.iter().take(num_event as usize) {
                if e.u64 == self.wayland_fd as u64 {
                    wayland_socket_ready = true;
                }
                if e.u64 == timer_fd as u64 {
                    timer_ready = true;
                }
            }
        }
        return Some((wayland_socket_ready, timer_ready));
    }

    /// Blocking read for the timer
    pub fn read_timer(&mut self) {
        self.timer.read();
    }
}
