use std::collections::VecDeque;

//in numeric representation (quadrants): 1132 , with 4 quadrants this corresponds to topleft, topright, bottomleft, bottomright
const DEFAULT_PASSCODE: [u8; 4] = [1, 1, 3, 2];

fn read_passcode_char(value: char, resolution: u8) -> Result<u8, &'static str> {
    match value {
        '0' if resolution == 4 => Ok(0),
        '1' => Ok(1),
        '2' if resolution >= 2 => Ok(2),
        '3' if resolution >= 2 => Ok(3),
        '4' if resolution >= 2 => Ok(4),
        '5' if resolution >= 3 => Ok(5),
        '6' if resolution >= 3 => Ok(6),
        '7' if resolution >= 3 => Ok(7),
        '8' if resolution >= 3 => Ok(8),
        '9' if resolution >= 3 => Ok(9),
        'A' | 'a' if resolution == 4 => Ok(10),
        'B' | 'b' if resolution == 4 => Ok(11),
        'C' | 'c' if resolution == 4 => Ok(12),
        'D' | 'd' if resolution == 4 => Ok(13),
        'E' | 'e' if resolution == 4 => Ok(14),
        'F' | 'f' if resolution == 4 => Ok(15),
        _ => {
            if resolution == 2 {
                Err("invalid quadrant for passcodeitem: 1 (topleft), 2 (topright), 3 (bottomleft), 4 (bottomright)")
            } else if resolution == 3 {
                Err("invalid quadrant for passcodeitem: use 1 (topleft),2,3 (topright),4,5 (center),6,7 (bottomleft),8,9 (bottomright)")
            } else if resolution == 4 {
                Err("invalid quadrant for passcodeitem: use 0 (topleft),1,2,3(topright),4,5,6,7,8,9,a,b,c (bottomleft),d,e,f (bottomright)")
            } else {
                Err("invalid quadrant for passcodeitem")
            }
        }
    }
}

#[derive(Clone, PartialEq, Debug)]
pub struct TouchState {
    passcode: Vec<u8>,
    touches: VecDeque<u8>,
    resolution: u8,
    max_touches: usize,
    grid_width: f64,
    grid_height: f64,
    tries: usize,
    max_tries: u16,
}

impl TouchState {
    pub fn new(
        width: i32,
        height: i32,
        passcode: Option<&str>,
        resolution: u8,
        max_tries: u16,
    ) -> Self {
        if resolution < 1 || resolution > 4 {
            panic!("Invalid resolution");
        }
        Self {
            passcode: if let Some(passcode) = passcode {
                passcode
                    .chars()
                    .map(|x| read_passcode_char(x, resolution).expect("Invalid passcode"))
                    .collect()
            } else if resolution == 1 {
                vec![1]
            } else {
                DEFAULT_PASSCODE.iter().copied().collect()
            },
            touches: VecDeque::new(),
            resolution,
            max_touches: if let Some(passcode) = passcode {
                passcode.chars().count()
            } else if resolution == 1 {
                1
            } else {
                DEFAULT_PASSCODE.len()
            },
            grid_width: width as f64 / (resolution as f64),
            grid_height: height as f64 / (resolution as f64),
            tries: 0,
            max_tries,
        }
    }

    pub fn resize(&mut self, width: i32, height: i32) {
        self.grid_width = width as f64 / (self.resolution as f64);
        self.grid_height = height as f64 / (self.resolution as f64);
    }

    pub fn add_touch(&mut self, x_val: f64, y_val: f64) -> u8 {
        let startindex: u8 = if self.resolution == 4 { 0 } else { 1 };

        let x = (x_val / self.grid_width).floor() as u8;
        let y = (y_val / self.grid_height).floor() as u8;
        let code = y * self.resolution + x + startindex;

        self.touches.push_back(code);
        while self.touches.len() > self.max_touches {
            self.touches.pop_front();
        }
        code
    }

    pub fn is_pass_valid(&self) -> bool {
        if self.permalocked() {
            false
        } else {
            self.touches == self.passcode
        }
    }

    pub fn fail(&mut self) {
        if self.touches.len() == self.passcode.len() && self.max_tries != 0 {
            self.tries += 1
        }
    }

    pub fn permalocked(&self) -> bool {
        self.max_tries != 0
            && self.tries >= std::cmp::max(self.passcode.len(), self.max_tries as usize)
    }

    pub fn reset(&mut self) {
        self.tries = 0;
    }
}
