use std::collections::VecDeque;
use std::sync::{Arc, Mutex};
use wayland_client::{
    delegate_noop,
    protocol::{
        wl_callback, wl_compositor,
        wl_keyboard::WlKeyboard,
        wl_output::{self, WlOutput},
        wl_registry, wl_seat, wl_surface,
        wl_touch::{self, WlTouch},
    },
    Connection, Dispatch, EventQueue, QueueHandle,
};

use wayland_client::protocol::wl_seat::WlSeat;
use wayland_protocols::ext::session_lock::v1::client::{
    ext_session_lock_manager_v1::{self, ExtSessionLockManagerV1},
    ext_session_lock_surface_v1, ext_session_lock_v1,
};
use wayland_protocols::{
    ext::session_lock::v1::client::ext_session_lock_v1::ExtSessionLockV1,
    xdg::xdg_output::zv1::client::zxdg_output_v1,
};

use wayland_protocols::xdg::shell::client::{xdg_toplevel, xdg_wm_base};
use wayland_protocols::xdg::xdg_output::zv1::client::zxdg_output_manager_v1;

use wayland_client::protocol::wl_buffer;
use wayland_client::protocol::wl_keyboard;
use wayland_client::protocol::wl_shm;
use wayland_client::protocol::wl_shm_pool;

use crate::touch_pass::TouchState;
use chrono::{DateTime, Local};

#[derive(Copy, Clone)]
pub struct TouchDown {
    pub serial: u32,
    pub x: f64,
    pub y: f64,
}

// This struct represents the state of our app. This simple app does not
// need any state, by this type still supports the `Dispatch` implementations.
pub struct State {
    lock: Option<ExtSessionLockV1>,
    lock_manager: Option<ExtSessionLockManagerV1>,
    pub connection: Connection,
    pub queue_handle: QueueHandle<State>,
    compositor: Option<wl_compositor::WlCompositor>,
    output: Option<WlOutput>,
    touch: Option<WlTouch>,
    keyboard: Option<WlKeyboard>,
    pub last_down: Option<TouchDown>,
    pub width: Option<i32>,
    pub height: Option<i32>,
    pub scale: i32,
    pub fontdesc_time: pango::FontDescription,
    pub fontdesc_status: pango::FontDescription,
    seat: Option<WlSeat>,
    pub buffers: VecDeque<wl_buffer::WlBuffer>,
    pub base_surface: Option<wl_surface::WlSurface>,
    pub inverted: bool,
    pub configured: bool,
    running: bool,
    pub redraw: bool,
    pub debug: bool, // In debug mode, a simple ESC suffices to unlock (useful for testing without touch support)
    notouch: bool,   // disable touch support  (for devices that don't have it)
    pub nofeedback: bool, // disable visual feedback
    wm_base: Option<xdg_wm_base::XdgWmBase>,
    zxdg_manager: Option<zxdg_output_manager_v1::ZxdgOutputManagerV1>,
    pub wl_shm: Option<wl_shm::WlShm>,
    pub touch_state: Option<TouchState>,
    pub status: Arc<Mutex<String>>,
    pub statushash: Option<u64>, //hash of last status
    pub time: Option<DateTime<Local>>,
}

impl State {
    /// Is the program still running?
    pub fn is_running(&self) -> bool {
        self.running
    }

    pub fn lock(&mut self) {
        let lock = self
            .lock_manager
            .as_ref()
            .unwrap()
            .lock(&self.queue_handle, ());

        self.lock = Some(lock);
        self.connection.roundtrip().expect("Failed to lock");
    }

    pub fn round(&self) {
        if let Err(e) = self.connection.roundtrip() {
            eprintln!("[peanutbutter] Failed roundtrip: {}", e);
        }
    }

    pub fn unlock(&mut self) {
        self.lock.as_ref().unwrap().unlock_and_destroy();
        self.lock = None;
        self.connection.roundtrip().expect("Failed to unlock");
        self.lock_manager.as_ref().unwrap().destroy();
        self.lock_manager = None;
    }

    /// Mark colours as inverted (triggered after touch)
    pub fn invert(&mut self) {
        self.inverted = !self.inverted;
    }
}

// Implement `Dispatch<WlRegistry, ()> for out state. This provides the logic
// to be able to process events for the wl_registry interface.
//
// The second type parameter is the user-data of our implementation. It is a
// mechanism that allows you to associate a value to each particular Wayland
// object, and allow different dispatching logic depending on the type of the
// associated value.
//
// In this example, we just use () as we don't have any value to associate. See
// the `Dispatch` documentation for more details about this.
impl Dispatch<wl_registry::WlRegistry, ()> for State {
    fn event(
        state: &mut Self,
        registry: &wl_registry::WlRegistry,
        event: wl_registry::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        // When receiving events from the wl_registry, we are only interested in the
        // `global` event, which signals a new available global.
        // When receiving this event, we just print its characteristics in this example.
        if let wl_registry::Event::Global {
            name,
            interface,
            version: _,
        } = event
        {
            match &interface[..] {
                "ext_session_lock_manager_v1" => {
                    let lock_manager =
                        registry.bind::<ExtSessionLockManagerV1, _, _>(name, 1, qh, ());
                    state.lock_manager = Some(lock_manager);
                }
                "wl_compositor" => {
                    let compositor =
                        registry.bind::<wl_compositor::WlCompositor, _, _>(name, 4, qh, ());
                    let surface = compositor.create_surface(qh, ());
                    state.base_surface = Some(surface);
                    state.compositor = Some(compositor);
                }
                "wl_output" => {
                    let output = registry.bind::<WlOutput, _, _>(name, 3, qh, ());
                    // Put first output as pin entry output
                    // TODO: allow user to set env variable to choose display
                    // Moreover, this code will cause bugs when new displays are plugged in
                    // Do we just want to display the pin entry on all displays? If so,
                    // use a vector
                    state.output = Some(output);
                }
                "wl_seat" => {
                    let seat = registry.bind::<WlSeat, _, _>(name, 1, qh, ());
                    // TODO: JUST SUPPORT TOUCH! DONT DO cursor https://github.com/Smithay/wayland-rs/issues/85
                    if !state.notouch {
                        state.touch = Some(seat.get_touch(qh, ()));
                    } else {
                        state.keyboard = Some(seat.get_keyboard(qh, ()));
                    }
                    state.seat = Some(seat);
                }
                "wl_shm" => {
                    let shm = registry.bind::<wl_shm::WlShm, _, _>(name, 1, qh, ());

                    state.wl_shm = Some(shm);
                }
                "xdg_wm_base" => {
                    let wm_base = registry.bind::<xdg_wm_base::XdgWmBase, _, _>(name, 1, qh, ());
                    state.wm_base = Some(wm_base);
                }
                "zxdg_output_manager_v1" => {
                    let zxdg = registry.bind::<zxdg_output_manager_v1::ZxdgOutputManagerV1, _, _>(
                        name,
                        1,
                        qh,
                        (),
                    );
                    state.zxdg_manager = Some(zxdg);
                }
                _ => {}
            }
        }
    }
}

impl Dispatch<wl_touch::WlTouch, ()> for State {
    fn event(
        state: &mut Self,
        _touch: &wl_touch::WlTouch,
        event: wl_touch::Event,
        _: &(),
        _: &Connection,
        _qh: &QueueHandle<Self>,
    ) {
        match event {
            wl_touch::Event::Down { serial, x, y, .. } => {
                state.last_down = Some(TouchDown { serial, x, y })
            }
            _ => {}
        }
    }
}

impl Dispatch<ext_session_lock_surface_v1::ExtSessionLockSurfaceV1, ()> for State {
    fn event(
        state: &mut Self,
        lock_surface: &ext_session_lock_surface_v1::ExtSessionLockSurfaceV1,
        event: ext_session_lock_surface_v1::Event,
        _: &(),
        _: &Connection,
        _qh: &QueueHandle<Self>,
    ) {
        match event {
            ext_session_lock_surface_v1::Event::Configure {
                serial,
                width,
                height,
            } => {
                state.configured = true;
                state.width = Some(width as i32);
                state.height = Some(height as i32);
                if state.debug {
                    eprintln!("ack configure: dimensions {width}x{height}");
                }
                lock_surface.ack_configure(serial);
                //forced redrawing as we may need to commit a new surface (if dimensions changed or it is the first)
                state.draw();
                state.base_surface.as_ref().unwrap().commit();
            }
            _ => unreachable!(),
        }
    }
}

impl Dispatch<ext_session_lock_v1::ExtSessionLockV1, ()> for State {
    fn event(
        state: &mut Self,
        _compositor: &ext_session_lock_v1::ExtSessionLockV1,
        event: ext_session_lock_v1::Event,
        _: &(),
        _: &Connection,
        qh: &QueueHandle<Self>,
    ) {
        match event {
            ext_session_lock_v1::Event::Locked => {
                let surface = state.compositor.as_ref().unwrap().create_surface(&qh, ());
                let _lock_surface = state.lock.as_ref().unwrap().get_lock_surface(
                    &surface,
                    state.output.as_ref().unwrap(),
                    &qh,
                    (),
                );
                surface.attach(state.buffers.iter().last(), 0, 0);
                state.base_surface = Some(surface);
            }
            _ => unreachable!(),
        }
    }
}

pub fn init_wayland(
    fontname: &str,
    debug: bool,
    notouch: bool,
    nofeedback: bool,
    status: Arc<Mutex<String>>,
) -> (State, EventQueue<State>) {
    // Create a Wayland connection by connecting to the server through the
    // environment-provided configuration.
    let conn = Connection::connect_to_env().unwrap();

    // Retrieve the WlDisplay Wayland object from the connection. This object is
    // the starting point of any Wayland program, from which all other objects will
    // be created.
    let display = conn.display();

    // Create an event queue for our event processing
    let mut event_queue = conn.new_event_queue();
    // An get its handle to associated new objects to it
    let qh = event_queue.handle();

    // Create a wl_registry object by sending the wl_display.get_registry request
    // This method takes two arguments: a handle to the queue the newly created
    // wl_registry will be assigned to, and the user-data that should be associated
    // with this registry (here it is () as we don't need user-data).
    let _registry = display.get_registry(&qh, ());

    // At this point everything is ready, and we just need to wait to receive the events
    // from the wl_registry, our callback will print the advertized globals.

    let mut fontdesc_time = pango::FontDescription::new();
    fontdesc_time.set_family(fontname);
    //size is configured in scale event

    let mut fontdesc_status = pango::FontDescription::new();
    fontdesc_status.set_family(fontname);
    //size is configured in scale event

    let mut state = State {
        lock: None,
        lock_manager: None,
        connection: conn,
        queue_handle: qh,
        compositor: None,
        output: None,
        touch: None,
        keyboard: None,
        seat: None,
        buffers: VecDeque::new(),
        base_surface: None,
        configured: false,
        redraw: false,
        inverted: false,
        fontdesc_time,
        fontdesc_status,
        wm_base: None,
        last_down: None,
        width: None,
        height: None,
        scale: 1, //will be set by wloutput scale callback
        debug,
        notouch,
        nofeedback,
        running: true,
        zxdg_manager: None,
        wl_shm: None,
        touch_state: None,
        status,
        statushash: None,
        time: None,
    };

    // To actually receive the events, we invoke the `sync_roundtrip` method. This method
    // is special and you will generally only invoke it during the setup of your program:
    // it will block until the server has received and processed all the messages you've
    // sent up to now.
    //
    // In our case, that means it'll block until the server has received our
    // wl_display.get_registry request, and as a reaction has sent us a batch of
    // wl_registry.global events.
    //
    // `sync_roundtrip` will then empty the internal buffer of the queue it has been invoked
    // on, and thus invoke our `Dispatch` implementation that prints the list of advertized
    // globals.
    event_queue.blocking_dispatch(&mut state).unwrap();

    let _output = state.zxdg_manager.as_ref().unwrap().get_xdg_output(
        state.output.as_ref().unwrap(),
        &state.queue_handle,
        (),
    ); // launch zxdg dispatch for setting width and height
    event_queue.blocking_dispatch(&mut state).unwrap();

    return (state, event_queue);
}

impl Dispatch<xdg_wm_base::XdgWmBase, ()> for State {
    fn event(
        _: &mut Self,
        wm_base: &xdg_wm_base::XdgWmBase,
        event: xdg_wm_base::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        if let xdg_wm_base::Event::Ping { serial } = event {
            wm_base.pong(serial);
        }
    }
}

impl Dispatch<xdg_toplevel::XdgToplevel, ()> for State {
    fn event(
        state: &mut Self,
        _: &xdg_toplevel::XdgToplevel,
        event: xdg_toplevel::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        if let xdg_toplevel::Event::Close {} = event {
            if state.debug {
                state.running = false;
            }
        }
    }
}

impl Dispatch<zxdg_output_v1::ZxdgOutputV1, ()> for State {
    fn event(
        state: &mut Self,
        _zxdg_output: &zxdg_output_v1::ZxdgOutputV1,
        event: zxdg_output_v1::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        match event {
            zxdg_output_v1::Event::LogicalSize { width, height } => {
                state.width = Some(width);
                state.height = Some(height);
                if let Some(touch_state) = state.touch_state.as_mut() {
                    touch_state.resize(width, height);
                }
                state.redraw = true;
                if state.debug {
                    eprintln!("logicalsize: dimensions {width}x{height}");
                }
            }
            _ => {}
        }
    }
}

impl Dispatch<wl_keyboard::WlKeyboard, ()> for State {
    fn event(
        state: &mut Self,
        _: &wl_keyboard::WlKeyboard,
        event: wl_keyboard::Event,
        _: &(),
        _: &Connection,
        _: &QueueHandle<Self>,
    ) {
        if let wl_keyboard::Event::Key { .. } = event {
            if state.debug {
                // any key stops the locker in debug mode
                state.running = false;
            }
        }
    }
}

impl Dispatch<wl_callback::WlCallback, wl_surface::WlSurface> for State {
    /// This will be triggered when the compositor deems it's a good time to draw a frame
    fn event(
        state: &mut Self,
        _proxy: &wl_callback::WlCallback,
        _event: <wl_callback::WlCallback as wayland_client::Proxy>::Event,
        _data: &wl_surface::WlSurface,
        _conn: &Connection,
        _qhandle: &QueueHandle<Self>,
    ) {
        if state.redraw {
            if state.debug {
                eprintln!("frame (redraw)");
            }
            state.draw();
            state.redraw = false;
            state.round();
        } else if state.debug {
            eprintln!("frame (redraw=false)");
        }
    }
}

impl Dispatch<wl_output::WlOutput, ()> for State {
    fn event(
        state: &mut Self,
        _touch: &wl_output::WlOutput,
        event: wl_output::Event,
        _: &(),
        _: &Connection,
        _qh: &QueueHandle<Self>,
    ) {
        match event {
            wl_output::Event::Scale { factor } => {
                if state.debug {
                    eprintln!("wloutput scaling factor = {factor}");
                }
                state.scale = factor;
                state.fontdesc_time.set_size(80 * pango::SCALE * factor);
                state.fontdesc_status.set_size(16 * pango::SCALE * factor);
            }
            _ => {}
        }
    }
}

delegate_noop!(State: ignore wl_shm::WlShm);
delegate_noop!(State: ignore wl_shm_pool::WlShmPool);
delegate_noop!(State: ignore wl_buffer::WlBuffer);
delegate_noop!(State: ignore ext_session_lock_manager_v1::ExtSessionLockManagerV1);
delegate_noop!(State: ignore wl_surface::WlSurface);
delegate_noop!(State: ignore wl_seat::WlSeat);
delegate_noop!(State: ignore wl_compositor::WlCompositor);
delegate_noop!(State: ignore zxdg_output_manager_v1::ZxdgOutputManagerV1);
